#ifndef STATEVECTORSSET_H
#define STATEVECTORSSET_H

#include <map>
//#include <atomic>
#include <iostream>
#include <cmath>
#include "particleset.h"
#include "label.h"
#include "ziggurat.h"
#include "prettyprintstl.h"

/// These variables govern the noise in the particle
/// resampling stage. We have not tested other values,
/// it probably does not actually belong here.
const float sqrtQCentreX = 5.477225575f; // sqrt(30)
const float sqrtQCentreY = 5.477225575f; // sqrt(30)
const float sqrtQSpeedX  = 1.0f;
const float sqrtQSpeedY  = 1.0f;
const float sqrtQRadiusX = 4.472135955f; // sqrt(20)
const float sqrtQRadiusY = 4.472135955f; // sqrt(20)

class StateVectorsSet {
public:
    struct StateVector : Particle {
        StateVector() : valid(false) {}
        StateVector(unsigned int id) : valid(true), id(id) {
            this->jsr = id;
            r4_nor_setup ( this->kn, this->fn, this->wn );
        }
        double squaredMahalanobisDistance (const PixelInRegion & px) const {
            const float pxRow = static_cast<float>(px.row);
            const float pxCol = static_cast<float>(px.col);
            assert(0.0 < radiusX);
            assert(0.0 < radiusY);
            const float relativeRowDistance = (pxRow - this->centreY) / radiusY;
            const float relativeColDistance = (pxCol - this->centreX) / radiusX;
            return relativeColDistance * relativeColDistance + relativeRowDistance * relativeRowDistance;
        }

        void setAssociatedLabel(unsigned int label) {
            this->associatedLabel = label;
            for (std::vector<Particle>::iterator p = this->associatedParticles.begin();
                 p != this->associatedParticles.end(); ++p) {
                p->associatedLabel = label;
            }
        }

        void redistributeParticles () {
            const unsigned int numParticles = this->associatedParticles.size();
            assert(0 < numParticles); // This doesn't make sense if there are no particles, it should never happen
            if (1 == numParticles) {
                return; // Nothing to distribute, the current "distribution" is good
            }
            assert(1.0f == this->weight);
            const float meanParticleWeight  = this->weight / static_cast<float>(numParticles);
            unsigned int allocated = 0;
            std::vector<unsigned int> votes;
            votes.assign(this->associatedParticles.size(), 0);
            std::vector<std::pair<float, unsigned int> > remainderIndex;
            for (unsigned int i = 0; i < this->associatedParticles.size(); ++i) {
                const float quota = this->associatedParticles.at(i).weight / meanParticleWeight;
                votes.at(i) = static_cast<unsigned int>(std::floor(quota));
                remainderIndex.push_back(std::make_pair<float, unsigned int>(quota - votes.at(i), i));
                allocated += votes.at(i);
            }
            const unsigned int surplus = this->associatedParticles.size() - allocated;
            if (0 < surplus) {
//                std::cout << remainderIndex << std::endl;
                std::nth_element(remainderIndex.begin(), remainderIndex.begin() + surplus - 1, remainderIndex.end(), std::greater<std::pair<float, unsigned int> >());
                for (unsigned int i = 0; i < surplus; ++i) {
                    ++votes.at(remainderIndex.at(i).second);
                }
            }
//            std::cout << "Surplus: " << surplus << std::endl;
//            std::cout << remainderIndex << std::endl;
//            std::cout << votes << std::endl;
            std::vector<Particle> redistributedParticles;
            redistributedParticles.reserve(this->associatedParticles.size());
            for (unsigned int i = 0; i < this->associatedParticles.size(); ++i) {
                for (unsigned int j = 0; j < votes.at(i); ++j) {
                    Particle p = this->associatedParticles.at(i);
                    p.weight = meanParticleWeight;
                    redistributedParticles.push_back(p);
                }
            }
            this->associatedParticles.swap(redistributedParticles);
        }

        void applyDynamicModel (float minRadius, float maxRadius) {
            this->centreX += this->speedX;
            this->centreY += this->speedY;
            for (std::vector<Particle>::iterator p = this->associatedParticles.begin(); p != this->associatedParticles.end(); ++p) {
                /// We force the first particle to be exactly the same as the state vector. This should avoid the case
                /// where the state vector covers a region but none of its associated particles do, getting zero weight
                if (this->associatedParticles.begin() == p) {
                    p->centreX = this->centreX;
                    p->centreY = this->centreY;
                    p->speedX = this->speedX;
                    p->speedY = this->speedY;
                    p->radiusX = this->radiusX;
                    p->radiusY = this->radiusY;
                } else {
                    p->centreX += p->speedX + r4_nor(this->jsr, this->kn, this->fn, this->wn) * sqrtQCentreX;
                    p->centreY += p->speedY + r4_nor(this->jsr, this->kn, this->fn, this->wn) * sqrtQCentreY;
                    p->speedX  +=             r4_nor(this->jsr, this->kn, this->fn, this->wn) * sqrtQSpeedX;
                    p->speedY  +=             r4_nor(this->jsr, this->kn, this->fn, this->wn) * sqrtQSpeedY;
                    const float newRadiusX = p->radiusX + r4_nor(this->jsr, this->kn, this->fn, this->wn) * sqrtQRadiusX;
                    const float newRadiusY = p->radiusY + r4_nor(this->jsr, this->kn, this->fn, this->wn) * sqrtQRadiusY;
                    p->radiusX = std::min(maxRadius, std::max(minRadius, newRadiusX));
                    p->radiusY = std::min(maxRadius, std::max(minRadius, newRadiusY));
                }
            }
        }








        //// Structures needed for the normal distribution RNG
        uint32_t jsr;
        uint32_t kn[128];
        float    fn[128];
        float    wn[128];
        //////////////////////////////////////////////////////


        bool valid;
        unsigned int id;
        std::vector<Particle> associatedParticles;
    };

    StateVectorsSet ()
        : initialized(false),lastId(0) {}

    void initializeParameters (unsigned int maxPart,
                               float minRadius,
                               float maxRadius) {
        assert(maxRadius >= minRadius);
        this->maxRadius = maxRadius;
        this->minRadius = minRadius;
        this->maxPart   = maxPart;
        this->initialized = true;
    }

    // This method must only be called after relabelling pixels with their associated State Vector IDs
    template <typename T, unsigned int CHANNELS>
    void movePixelsIntoTheFuture (std::vector<ColouredPixel<T, CHANNELS> > & pixels) const {
        assert(this->initialized);
        for (typename std::vector<ColouredPixel<T, CHANNELS> >::iterator px = pixels.begin(); px != pixels.end(); ++px) {
            std::map<unsigned int, StateVectorsSet::StateVector>::const_iterator sv = this->data.find(px->label);
            if (sv != this->data.end()) {
                const StateVectorsSet::StateVector & associatedStateVector = sv->second;
                px->row += lrint(associatedStateVector.speedY);
                px->col += lrint(associatedStateVector.speedX);
            }
        }
    }




    void clearCurrentRegions () {
        assert(this->initialized);
        this->data.clear();
    }

    /// TODO : Note that this is NOT thread-safe. We cannot use atomic types from C++ 11 because
    /// we are still using CUDA 4.2, which does not support them. Actually, we could but not in
    /// code that must be reachable from both host and device
    unsigned int createStateVector (const RegionInfo & regionInfo) {
        assert(this->initialized);
//        unsigned int mostNumerousSV = 0;
//        unsigned int numberOfParticles = 0;
//        for (std::map<unsigned int, StateVector>::const_iterator it = this->data.begin(); it != this->data.end(); ++it  ) {
//            if (it->second.associatedParticles.size() > numberOfParticles) {
//                mostNumerousSV = it->first;
//                numberOfParticles = it->second.associatedParticles.size();
//            }
//        }

//        std::cout << "Most numerous SV: " << mostNumerousSV << std::endl;
        // FIXME : ¿Esto ha desaparecido de aquí porque lo hacemos luego con la redistribución por tamaños, creo?
        /// Quitamos la partícula de menor peso (TODO, de momento la última, ya veremos dónde lo ordenamos)
        /// para hacer sitio al vector de estado nuevo
        //assert(1 < numberOfParticles);
        //this->data[mostNumerousSV].associatedParticles.pop_back();


        const unsigned int vectorId = ++this->lastId; // First [non-atomic] increment, then return value

        StateVector newStateVector(vectorId);
        newStateVector.centreX = regionInfo.medianCol;
        newStateVector.centreY = regionInfo.medianRow;
        newStateVector.radiusX = std::min(this->maxRadius, std::max(this->minRadius, (regionInfo.maxCol - regionInfo.minCol) / 2.0f)); // minParticleSize MUST be > 0
        newStateVector.radiusY = std::min(this->maxRadius, std::max(this->minRadius, (regionInfo.maxRow - regionInfo.minRow) / 2.0f)); // minParticleSize MUST be > 0
        newStateVector.speedX  = 0.0f;
        newStateVector.speedY  = 0.0f;
        newStateVector.associatedLabel = regionInfo.label;
        newStateVector.weight = 1.0f;

        Particle part = newStateVector; // Slicing occurs, but this is OK
        assert(0.0f < part.radiusX);
        assert(0.0f < part.radiusY);

        newStateVector.associatedParticles.push_back(part); ////////////


        this->data.insert(std::pair<unsigned int, StateVector>(vectorId, newStateVector));






//        for (std::map<unsigned int, StateVector>::const_iterator it = this->data.begin(); it != this->data.end(); ++it  ) {
//            std::cout << it->first << " => " << it->second.associatedParticles.size() << '\n';
//        }

//        /// TODO : comprobar que lo que hemos hecho aquí tiene algo de sentido
//        asdfasdfa




        return vectorId;
    }

    ParticleManager getAggregateParticles () {
        assert(this->initialized);
        ParticleManager pm(this->maxPart);
        for (std::map<unsigned int, StateVector>::const_iterator it = this->data.begin(); it != this->data.end(); ++it  ) {
            pm.appendParticles(it->second.associatedParticles);
        }
        return pm;
    }

    ParticleManager getAggregateStateVectorsAsParticles () {
        assert(this->initialized);
        ParticleManager pm(this->maxPart);
        std::vector<Particle> stateVectorsAsParticles;
        for (std::map<unsigned int, StateVector>::const_iterator it = this->data.begin(); it != this->data.end(); ++it  ) {
            stateVectorsAsParticles.push_back(it->second);
        }
        pm.appendParticles(stateVectorsAsParticles);
        return pm;
    }

    bool initialized;
    unsigned int maxPart;
    float minRadius;
    float maxRadius;

    std::map<unsigned int, StateVector> data; // id del SV -> SV
    unsigned int lastId;
};

#endif // STATEVECTORSSET_H
