#include <cstdlib>
#include <cuda_runtime.h>
#include <cstring>
#include <cassert>
#include <iostream>
#include <FreeImagePlus.h>
#include "auxcuda.h"
#include "thumbnail.h"


Thumbnail::Thumbnail()
    : width(0), height(0), hostData(NULL), auxHostData(NULL), deviceData(NULL), enabled(false), frame(1), iteration(0)
{ }


Thumbnail::Thumbnail(unsigned int width, unsigned int height, unsigned int seed)
    : width(0), height(0), hostData(NULL), auxHostData(NULL), deviceData(NULL), enabled(false), frame(1), iteration(0)
{
    this->allocate(width, height);
    tinymt32_init(&this->rngState, seed);
    std::cout << "Construyendo thumbnail" << std::endl;
}

void Thumbnail::allocate (unsigned int width, unsigned int height) {
    assert(NULL == this->hostData);
    assert(NULL == this->auxHostData);
    assert(NULL == this->deviceData);
    this->width  = width;
    this->height = height;
    CUDA_SAFE_CALL (cudaMallocHost (&this->hostData, sizeof (unsigned char) * width * height));
    CUDA_SAFE_CALL (cudaMallocHost (&this->auxHostData, sizeof (unsigned char) * width * height));
    CUDA_SAFE_CALL (cudaMalloc (&this->deviceData, sizeof (unsigned char) * width * height));
}

void Thumbnail::setRandomSeed (unsigned int seed) {
    this->rngState.mat1 = seed;
    this->rngState.mat2 = seed;
    this->rngState.tmat = seed;
    this->rngState.status[0] = seed;
    this->rngState.status[1] = seed;
    this->rngState.status[2] = seed;
    this->rngState.status[3] = seed;
    tinymt32_init(&this->rngState, seed);
}

bool Thumbnail::writeImage() {
    if (!this->enabled) {
        return true;
    }
    static fipImage out;
    out.setSize(FIT_BITMAP, this->width, this->height, 24);
    for (unsigned int r = 0; r < this->height; ++r) {
        unsigned char * ptr = out.getScanLine(r);
        for (unsigned int c = 0; c < this->width; ++c) {
            unsigned char blue, green, red;
            switch (this->elementAt(c, r)) {
            case NOT_ANALYSED:
                blue  = 0;
                green = 0;
                red   = 0;
                break;
            case TO_ANALYSE:
                blue  = 255;
                green = 0;
                red   = 0;
                break;
            case POSITIVE:
                blue  = 0;
                green = 255;
                red   = 0;
                break;
            case NEGATIVE:
                blue  = 0;
                green = 0;
                red   = 255;
                break;
            default:
                assert(false && "Invalid value in thumbnail!!!");
            }
            *ptr = blue;
            ++ptr;
            *ptr = green;
            ++ptr;
            *ptr = red;
            ++ptr;
        }
    }
    char filename[255];
    sprintf(filename, "/run/shm/thumb/%04d-%02d.png", this->frame, this->iteration);
    return out.save(filename);
}

void Thumbnail::release () {
    CUDA_SAFE_CALL (cudaFreeHost (this->hostData));
    this->hostData = NULL;
    CUDA_SAFE_CALL (cudaFreeHost (this->auxHostData));
    this->auxHostData = NULL;
    CUDA_SAFE_CALL (cudaFree (this->deviceData));
    this->deviceData = NULL;
    this->setEnabled(false);
}



bool Thumbnail::getEnabled() const
{
    return enabled;
}

void Thumbnail::setEnabled(bool value)
{
    enabled = value;
}


unsigned char *Thumbnail::getDeviceData()
{
    return deviceData;
}


unsigned char const * Thumbnail::getHostData() const
{
    return hostData;
}

unsigned char * Thumbnail::getHostData()
{
    return hostData;
}


unsigned int Thumbnail::getHeight() const
{
    return height;
}

unsigned int Thumbnail::getWidth() const
{
    return width;
}


Thumbnail::~Thumbnail() {
    this->release();
}

void Thumbnail::clear() {
    if (this->enabled) {
        memset (this->hostData, 0, this->width * this->height);
    }
}

unsigned char &Thumbnail::elementAt(int x, int y) {
    unsigned int const index1D = x + y * this->width;
    return this->hostData[index1D];
}

void Thumbnail::initializeNextFrame() {
    if (this->enabled) {
        for (unsigned int y = 0; y < height; ++y) {
            for (unsigned int x = 0; x < width; ++x) {
                unsigned char &element = this->elementAt (x, y);
                element = (POSITIVE == element) ? TO_ANALYSE : NOT_ANALYSED;
            }
        }
    }
}

void Thumbnail::addRandom(unsigned int blocksPerTile, unsigned int tileWidth, unsigned int tileHeight) {
    if (this->enabled) {
        for (unsigned int baseY = 0; baseY < this->height; baseY += tileHeight) {
            for (unsigned int baseX = 0; baseX < this->width; baseX += tileWidth) {
                for (unsigned int n = 0; n < blocksPerTile; ++n) {
                    unsigned int const offsetY = tinymt32_generate_uint32(&this->rngState) % tileHeight;
                    unsigned int const offsetX = tinymt32_generate_uint32(&this->rngState) % tileWidth;
                    // FIXME : Esta comprobación la he puesto pero no la he probado.
                    if (((offsetX + baseX) < this->width) && ((offsetY + baseY) < this->height)) {
                        unsigned int const index1D = (offsetX + baseX)
                                                     + (offsetY + baseY) * width;
                        this->hostData[index1D] = TO_ANALYSE;
                    }
                }
            }
        }
    }
}

bool Thumbnail::nextIteration(int radius) {
    if ((!this->enabled) || (0 >= radius)) {
        return false;
    }
    bool nextIterationNeeded = false;

    const int intHeight = this->height;
    const int intWidth  = this->width;

    /// We mark to analyse any block that is next (8-conn) to a positive
    for (int y = 0; y < intHeight; ++y) {
        for (int x = 0; x < intWidth; ++x) {
            unsigned char &currentElement = this->elementAt (x, y);
            this->auxHostData[x + y * intWidth] = NOT_ANALYSED;
            if (NOT_ANALYSED == currentElement) {
                for (int readY = std::max(0, y - 1); readY < std::min(intHeight, y + 2); ++readY) {
                    for (int readX = std::max(0, x - 1); readX < std::min(intWidth, x + 2); ++readX) {
                        if (POSITIVE == this->elementAt (readX, readY)) {
                            //currentElement = TO_ANALYSE;
                            this->auxHostData[x + y * intWidth] = TO_ANALYSE;
                            nextIterationNeeded = true;
                            goto firstLoopEnd; // We do not need to analyse
                            // the other neighbours, one is enough
                        }
                    }
                }
            }
            firstLoopEnd:
            ;
        }
    }


    /// We mark to analyse any block that is within radius-1 of a "to analyse".
    /// We do this in two phases because if we simply mark to analyse every block
    /// within radius of a positive, we overanalyse the edges of the object, while
    /// we merely want to converge faster in those areas of the object that are
    /// bigger than anticipated.
    for (int y = 0; y < intHeight; ++y) {
        for (int x = 0; x < intWidth; ++x) {
            unsigned char &currentElement = this->elementAt (x, y);
            if (NOT_ANALYSED == currentElement) {
                for (int readY = std::max(0, y - (radius - 1)); readY < std::min(intHeight, y + radius); ++readY) {
                    for (int readX = std::max(0, x - (radius - 1)); readX < std::min(intWidth, x + radius); ++readX) {
                        if (TO_ANALYSE == this->auxHostData[readX + readY * intWidth]) { //this->elementAt (readX, readY)) {
                            currentElement = TO_ANALYSE;
                            goto secondLoopEnd; // We do not need to analyse
                            // the other neighbours, one is enough
                        }
                    }
                }
            }
            secondLoopEnd:
            ;
        }
    }
    this->iteration++; // We advance this unconditionally so that we can see the final state
    return nextIterationNeeded;
}

void Thumbnail::toDevice(cudaStream_t stream) const {
    if (!this->enabled) {
        return;
    }
    if (NULL == stream) {
        CUDA_SAFE_CALL (cudaMemcpy (this->deviceData, this->hostData, sizeof (unsigned char) * width * height, cudaMemcpyHostToDevice));
    } else {
        CUDA_SAFE_CALL (cudaMemcpyAsync (this->deviceData, this->hostData, sizeof (unsigned char) * width * height, cudaMemcpyHostToDevice, stream));
    }
}

void Thumbnail::fromDevice(cudaStream_t stream) {
    if (!this->enabled) {
        return;
    }
    if (NULL == stream) {
        CUDA_SAFE_CALL (cudaMemcpy (this->hostData, this->deviceData, sizeof (unsigned char) * width * height, cudaMemcpyDeviceToHost));
    } else {
        CUDA_SAFE_CALL (cudaMemcpyAsync (this->hostData, this->deviceData, sizeof (unsigned char) * width * height, cudaMemcpyDeviceToHost, stream));
    }
}
